//This file is part of Furius Iso Mount. Copyright 2008 Dean Harris (marcus_furius@hotmail.com)
//
//    Furius Iso Mount is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    Furius Iso Mount is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with Furius Iso Mount.  If not, see <http://www.gnu.org/licenses/>.
using System;
using Gtk;
using System.IO;
using FuriusIsoMount;
using System.Threading;
using System.Diagnostics;
using System.Collections;
using System.Xml.Serialization;

public partial class MainWindow: Gtk.Window
{	
	private TreeStore isoListStore;
	private TreeSelection selection;
	private TreeIter iter;
	
	private Thread computeHashThread;
	
	private string[] history = new string[Globals.ImageHistoryMaximum];
	
	private bool isSelectedMountedImageFuse;
		
	/// <summary>
    /// Initializes a new instance of the <see cref="MainWindow"/> class.
    /// </summary>
	public MainWindow (): base (Gtk.WindowType.Toplevel)
	{
				
		CreateApplicationDirectory();
		Log.Write(Globals.MountLog, LocalizableResources.ApplicationStart);
		
		Build ();
		
		CreateIconsAndText();		
		CreateTreeviewColumns();
		ModifyLabelFontSizes();	
		//Set some tooltips
		SetControlToolTips();
		//Mount any previous images
		MountPreviousImages();
		//Load History
		LoadHistory();
	}
	
	/// <summary>
    /// Creates the Application directory to store the Logs and mount lists in
    /// /home/username/.FuriusIsoMount
    /// </summary>	
	private void CreateApplicationDirectory()
	{
		if (!Directory.Exists(Environment.GetFolderPath(Environment.SpecialFolder.Personal) +  
				                       "/.FuriusIsoMount/"))
		{
			try
			{
				Directory.CreateDirectory(Environment.GetFolderPath(Environment.SpecialFolder.Personal) +  
				                       "/.FuriusIsoMount/");
			}
			catch (IOException exception)
			{
				MessageBox.Show(this, exception.Message, MessageType.Error);
			}
			
		}
	}
	
	/// <summary>
	/// Loads text from the resource files
	/// </summary>
	private void CreateIconsAndText()
	{

		this.Title = string.Format("{0} ({1})", 
		                           AboutBox.ApplicationTitle, AboutBox.ApplicationVersion);
		
		//Create Button Text and Images
		this.buttonBrowse.Label = LocalizableResources.BrowseLabel;
		Gtk.Image buttonBrowseImage = new Gtk.Image();
		buttonBrowseImage.Pixbuf = Stetic.IconLoader.LoadIcon(this, "gtk-open", Gtk.IconSize.Menu, 16);
		this.buttonBrowse.Image = buttonBrowseImage;
		
		this.buttonMount.Label = LocalizableResources.MountLabel;
		Gtk.Image buttonMountImage = new Gtk.Image();
		buttonMountImage.Pixbuf = Stetic.IconLoader.LoadIcon(this, "gtk-harddisk", Gtk.IconSize.Menu, 16);
		this.buttonMount.Image = buttonMountImage;
		
		this.buttonChecksum.Label = LocalizableResources.ChecksumLabel;
		Gtk.Image buttonCheckImage = new Gtk.Image();
		buttonCheckImage.Pixbuf = Gdk.Pixbuf.LoadFromResource("imageChecksum.png");
		this.buttonChecksum.Image = buttonCheckImage;
		
		this.buttonBurn.Label = LocalizableResources.BurnLabel;
		Gtk.Image buttonBurnImage = new Gtk.Image();
		buttonBurnImage.Pixbuf = Gdk.Pixbuf.LoadFromResource("imageBurn.png");
		this.buttonBurn.Image = buttonBurnImage;
		
		this.buttonEventLog.Label = LocalizableResources.ViewLogLabel;
		this.buttonDeleteLog.Label = LocalizableResources.DeleteLogLabel;
		
		this.buttonUnmount.Label = LocalizableResources.UnMountLabel;
		Gtk.Image buttonUnMountImage = new Gtk.Image();
		buttonUnMountImage.Pixbuf = Stetic.IconLoader.LoadIcon(this, "gtk-cancel", Gtk.IconSize.Menu, 16);
		this.buttonUnmount.Image = buttonUnMountImage;
		
		this.labelSelectImage.Text = LocalizableResources.SelectImageLabel;
		this.labelSelectedImage.Text = LocalizableResources.SelectedImageLabel;
		this.labelHash.Text = LocalizableResources.HashLabel;
		this.labelSelectedMountPoint.Text = LocalizableResources.SelectedMountPointLabel;
		this.GtkLabelMountedImages.Text = LocalizableResources.MountedImagesHeading;
		this.GtkLabelMountedImages.UseMarkup = true;
	}
	
	/// <summary>
    /// Changes some labels to use a smaller 8 point font
	/// </summary>
	private void ModifyLabelFontSizes()
	{
		//Modify Font size
		this.labelSelectedImage.ModifyFont(Pango.FontDescription.FromString("8"));
		this.labelSelectedMountPoint.ModifyFont(Pango.FontDescription.FromString("8"));
		this.labelHash.ModifyFont(Pango.FontDescription.FromString("8"));
	}
	
	/// <summary>
    /// Sets the controls with localized tooltips
	/// </summary>
	private void SetControlToolTips()
	{	
		this.treeviewMountedImages.TooltipText = LocalizableResources.MountedImagesTooltip;
		this.buttonBrowse.TooltipText = LocalizableResources.BrowseButtonTooltip;
		this.buttonMount.TooltipText = LocalizableResources.MountButtonTooltip;
		this.buttonUnmount.TooltipText = LocalizableResources.UnmountButtonTooltip;
		this.buttonChecksum.TooltipText = LocalizableResources.ChecksumButtonTooltip;
		this.buttonBurn.TooltipText = LocalizableResources.BurnButtonTooltip;
	}
	
	/// <summary>
    /// Checks a csv file for previously unmounted images
	/// and attempts to add them to the mount list.
    /// </summary>
	private void MountPreviousImages()
	{
		//Try to load any mount points
		if (File.Exists(Globals.MountList))
		{
			try
			{
				using (StreamReader streamReader = new System.IO.StreamReader(Globals.MountList))
				{
					string line;
					while((line = streamReader.ReadLine()) != null)
					{
					    string[] mountPoint = line.Split(',');
						//Ensure the directory exists
						if (Directory.Exists(mountPoint[0]))
					    {
							if (mountPoint.Length < 3)
							{
								//This may be an old history, so assume fuse was used
								this.isoListStore.AppendValues (mountPoint[0], mountPoint[1], "True");
							}
							else
							{
								this.isoListStore.AppendValues (mountPoint[0], mountPoint[1], mountPoint[2]);
							}
						}
						else
						{
							Log.Write(Globals.MountLog, string.Format(LocalizableResources.NoExistMountPoint, mountPoint[0]));
						}
					}
				}
				// we've loaded the records to kill the file
				File.Delete(Globals.MountList);
			}
			catch(IOException exception)
			{
				Log.Write(Globals.MountLog, exception.Message);
			}
			catch(Exception exception)
			{
				Console.WriteLine(exception.ToString());
				Log.Write(Globals.MountLog, exception.Message);
			}
			
		}
	}
	
	/// <summary>
	/// Load the history
	/// </summary>
	private void LoadHistory()
	{
		if (File.Exists(Globals.HistoryList))
		{
			try
			{
				using (StreamReader streamReader = new System.IO.StreamReader(Globals.HistoryList))
				{
					string line;
					int i = 0;
					while((line = streamReader.ReadLine()) != null)
					{
					    this.entryImage.AppendText(line);
						Globals.InsertIntoArray(history, line, i);
						i++;
					}
				}
			}
			catch(IOException exception)
			{
				Log.Write(Globals.MountLog, exception.Message);
			}
			catch(Exception exception)
			{
				Log.Write(Globals.MountLog, exception.Message);
			}
			
		}
	}
		
	/// <summary>
	/// Creates the TreeView Columns and sets the datasource (TreeStore)
	/// </summary>
	private void CreateTreeviewColumns()
	{	 
		this.treeviewMountedImages.AppendColumn (LocalizableResources.MountPointColumnHeading, 
		                                         new Gtk.CellRendererText (), "text", 0);
		this.treeviewMountedImages.AppendColumn (LocalizableResources.ImageFileColumnHeading, 
		                                         new Gtk.CellRendererText (), "text", 1);
		this.treeviewMountedImages.AppendColumn ("Fuse", 
		                                         new Gtk.CellRendererText (), "text", 2);
		this.isoListStore = new Gtk.TreeStore (typeof (string), typeof (string), typeof (string));		
		this.treeviewMountedImages.Columns[0].SortColumnId = 0;
		this.treeviewMountedImages.Columns[1].SortColumnId = 1;
		this.treeviewMountedImages.Columns[2].SortColumnId = 2;
		this.treeviewMountedImages.Model = this.isoListStore;
				
	}
	
	/// <summary>
    /// Handles the DeleteEvent event of the MainWindow control.
    /// </summary>
    /// <param name="sender">The source of the event.</param>
    /// <param name="a">The <see cref="DeleteEventArgs"/> instance containing the event data.</param>
	protected void OnDeleteEvent (object sender, DeleteEventArgs a)
	{	

		this.treeviewMountedImages.Model.Foreach(CleanUp);
				
		Application.Quit ();
		a.RetVal = true;
		Log.Write(Globals.MountLog, LocalizableResources.ApplicationClosed);
		
	}
	
	/// <summary>
    /// Iterates over each TreeIter and request confirmation to unmount any mounted images
    /// </summary>
    /// <param name="model">TreeModel</param>
    /// <param name="path">TreePath</param>
	/// <param name="iter">TreeIter</param>
	/// <remarks>
	/// The three parameters: model, path, and iter, must be included when
    /// you define a function called by TreeModel.Foreach, but you don't have to specifically pass them
    /// because they will get pass automaticallly.
	/// If you return [true] Iteration cancels
	/// </remarks>
	/// <returns>[false]</returns>
	private bool CleanUp(TreeModel model, TreePath path, TreeIter iter)
	{
		
		if (MessageBox.Show(this, string.Format(LocalizableResources.ImageStillMounted, model.GetValue(iter, 1).ToString()), 
		                    MessageType.Question, ButtonsType.YesNo) == ResponseType.Yes)
		{
			try // and unmount the image
			{
				bool isFuse;
				try
				{
					isFuse = Convert.ToBoolean(model.GetValue(iter, 2));
				}
				catch (System.FormatException exception)
				{
					//Keep going... just assume that it's fuse
					isFuse = true;
					Log.Write(Globals.MountLog, string.Format("{0}{1}{2}", exception.Message, 
					                                          Environment.NewLine, exception.ToString()));
					
				}
				ImageAction.Unmount(model.GetValue(iter, 0).ToString(), this.isoListStore, this.iter, true, isFuse);
			}
			catch (Exception exception)
			{
				Log.Write(Globals.MountLog, exception.Message);
			}
		}
		else // Save to file for loading later
		{
			try 
			{
				using (StreamWriter writer = File.AppendText(Globals.MountList))
				{
					//Write a mountpoint
					writer.WriteLine(string.Format("{0},{1},{2}", model.GetValue(iter, 0).ToString(), 
					                           model.GetValue(iter, 1).ToString(),
					                               model.GetValue(iter, 2).ToString()));
				}
			}
			catch(IOException exception)
			{
				Log.Write(Globals.MountLog, exception.Message);
			}
		}
		
		return false;
	}
		
	/// <summary>
    /// Handles the Click event of the buttonBrowse control.
    /// </summary>
    /// <param name="sender">The source of the event.</param>
    /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
	protected virtual void OnButtonBrowseClicked (object sender, System.EventArgs e)
	{
		SelectAndVerifyImage();
	}
	
	/// <summary>
    /// Selects and Verifies the Image file by calling
	/// SelectImage and VerifyImage methods
    /// </summary>
	private void SelectAndVerifyImage()
	{		
		if(!SelectImage())
		{
			return;
		}
		VerifyImage();		
	}
	
	/// <summary>
    /// Handles the Click event of the buttonBrowse control.
    /// </summary>
	/// <returns>[true] if a file is selected</returns>
	private bool SelectImage()
	{
		// Create a new FileSelection Dialog
		using (FileChooserDialog fileSelection = new FileChooserDialog(LocalizableResources.SelectImageLabel, 
		                                                               this, FileChooserAction.Open, 
		                                                               LocalizableResources.CancelLabel,
		                                                               ResponseType.Cancel,
		                                                               LocalizableResources.OpenLabel, 
		                                                               ResponseType.Accept))
		{		
			if ( (Gtk.ResponseType) fileSelection.Run () != Gtk.ResponseType.Accept) 
			{
				fileSelection.Hide ();
				return false;
			}
			this.entryImage.Entry.Text = fileSelection.Filename;
			fileSelection.Hide ();
			return true;
		}
	}

	/// <summary>
    /// Verifies that an Image (.iso or .img) file has been selected.
    /// </summary>
	private void VerifyImage()
	{
		string file = this.entryImage.Entry.Text;
		// Check the file exists
		if (!File.Exists(file))
		{
			MessageBox.Show(this, LocalizableResources.FileNotFound, MessageType.Error);
			this.labelSelectedImage.Text = LocalizableResources.SelectedImageLabel;
			this.buttonMount.Sensitive = false;
			this.buttonChecksum.Sensitive = false;
			this.buttonBurn.Sensitive = false;
			return;
		}
		
		//Check the files an Image
		if (file.ToLower().EndsWith(".iso") || file.ToLower().EndsWith(".img") || file.ToLower().EndsWith(".bin")
		    || file.ToLower().EndsWith(".mdf") || file.ToLower().EndsWith(".nrg"))
		{
			this.labelSelectedImage.Text = file;
			this.buttonMount.Sensitive = true;
			this.buttonChecksum.Sensitive = true;
			this.buttonBurn.Sensitive = true;
			// Loop method only supports iso and img files
			if (file.ToLower().EndsWith(".iso") || file.ToLower().EndsWith(".img"))
			{
				this.radiobuttonLoop.Sensitive = true;
			}
			else
			{
				this.radiobuttonLoop.Sensitive = false;
				this.radiobuttonFuse.Active = true;
			}
			return;
		}

		MessageBox.Show(this, LocalizableResources.IncompatibleFile, MessageType.Error);
		this.labelSelectedImage.Text = LocalizableResources.SelectedImageLabel;
		this.buttonMount.Sensitive = false;
		this.buttonChecksum.Sensitive = false;
		this.buttonBurn.Sensitive = false;
				
	}

	/// <summary>
    /// Handles the Click event of the buttonMount control.
    /// </summary>
    /// <param name="sender">The source of the event.</param>
    /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
	protected virtual void OnButtonMountClicked (object sender, System.EventArgs e)
	{
		try // and mount the Image
		{
			ImageAction.Mount(this.labelSelectedImage.Text, this.isoListStore, 
			                  this.entryImage, this.history, this.radiobuttonFuse.Active);
		}
		catch (UnauthorizedAccessException exception)
		{
			MessageBox.Show(this, exception.Message, 
				                MessageType.Error);
		}
		catch (IOException exception)
		{
			MessageBox.Show(this, exception.Message +
				                LocalizableResources.MountError, 
				                MessageType.Error);
		}
	}

	/// <summary>
    /// Handles the CursorChanged event of the treeviewMountedImages control.
    /// </summary>
    /// <param name="sender">The source of the event.</param>
    /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
	protected virtual void OnTreeviewMountedImagesCursorChanged (object sender, System.EventArgs e)
	{
		TreeModel model;
		// Get the selected TreeModel and TreeIter
		this.selection = (sender as TreeView).Selection;
		selection.GetSelected(out model, out iter);
		//Get the value of the first item (0 based index) of the selected TreeIter
		this.labelSelectedMountPoint.Text = this.treeviewMountedImages.Model.GetValue(iter, 0).ToString();
		try
		{
			this.isSelectedMountedImageFuse = Convert.ToBoolean(this.treeviewMountedImages.Model.GetValue(iter, 2));
		}
		catch (System.FormatException exception)
		{
			//Keep going... just assume that it's fuse
			this.isSelectedMountedImageFuse = true;
			Log.Write(Globals.MountLog, string.Format("{0}{1}{2}", exception.Message, 
			                                          Environment.NewLine, exception.ToString()));
			
		}
		// Allow user to unmount
		this.buttonUnmount.Sensitive = true;
		
	}

	/// <summary>
    /// Handles the Click event of the buttonUnmount control.
    /// </summary>
    /// <param name="sender">The source of the event.</param>
    /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
	protected virtual void OnButtonUnmountClicked (object sender, System.EventArgs e)
	{
		try // and unmount the image
		{
			ImageAction.Unmount(this.labelSelectedMountPoint.Text, this.isoListStore, this.iter, false, this.isSelectedMountedImageFuse);
			//This will delete the current TreeIter so no mount points are selected.
			this.labelSelectedMountPoint.Text = LocalizableResources.SelectedMountPointLabel;
			this.buttonUnmount.Sensitive = false;
		}
		catch (UnauthorizedAccessException exception)
		{
			MessageBox.Show(this, exception.Message, 
				                MessageType.Error);
		}
		catch (IOException exception)
		{
			MessageBox.Show(this, exception.Message +
				                LocalizableResources.UnmountError, 
				                MessageType.Error);
		}
	}
	
	

	/// <summary>
    /// Handles the Click event of the buttonChecksum control.
    /// </summary>
    /// <param name="sender">The source of the event.</param>
    /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
	protected virtual void OnButtonChecksumClicked (object sender, System.EventArgs e)
	{	
		
		if (this.buttonChecksum.Label == LocalizableResources.ChecksumLabel)
		{
			this.computeHashThread = new Thread (new ThreadStart (ComputeChecksum));
			this.computeHashThread.Start ();
			Log.Write(Globals.MountLog, LocalizableResources.ChecksumStarted);
			this.labelHash.Text = LocalizableResources.CalculatingChecksum;
			this.buttonChecksum.Label = LocalizableResources.CancelLabel;
		}
		else
		{
			if (this.computeHashThread.IsAlive)
			{
				this.computeHashThread.Abort(); // Not the ideal way to stop a thread!
				this.labelHash.Text = LocalizableResources.HashLabel;
			}
			this.buttonChecksum.Label = LocalizableResources.ChecksumLabel;
			Log.Write(Globals.MountLog, LocalizableResources.ChecksumAborted);
		}
	}
	
	/// <summary>
    /// Computes the selected Checksum
    /// </summary>
	/// <remarks>
	/// Gtk is not thread safe, so uses Gtk.Application.Invoke
	/// to make code run on the main loop thread.
	/// </remarks>
	private void ComputeChecksum ()
    {
        string hashValue;
		if (this.radiobuttonMd5.Active)
		{
			hashValue = FileHash.ComputeHash(this.entryImage.Entry.Text, FileHash.Provider.MD5);
		}
		else
		{
			hashValue = FileHash.ComputeHash(this.entryImage.Entry.Text, FileHash.Provider.SHA1);
		}
		
		Gtk.Application.Invoke (delegate {
            this.labelHash.Text = hashValue;
			this.buttonChecksum.Label = LocalizableResources.ChecksumLabel;
			Log.Write(Globals.MountLog, LocalizableResources.ChecksumComplete 
			          + Environment.NewLine 
			          + this.labelSelectedImage.Text + ":" + this.labelHash.Text);
        });
    }

	/// <summary>
    /// Handles the Click event of the buttonBurn control.
    /// </summary>
    /// <param name="sender">The source of the event.</param>
    /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
	protected virtual void OnButtonBurnClicked (object sender, System.EventArgs e)
	{
		if (this.entryImage.Entry.Text.ToLower().EndsWith(".iso") 
		    || this.entryImage.Entry.Text.ToLower().EndsWith(".img"))
		{
			if (this.radiobuttonBrasero.Active)
			{
				ImageAction.BurnBrasero(this.entryImage.Entry.Text);
			}
			else
			{
				ImageAction.BurnNautilus(this.entryImage.Entry.Text);
			}
		}
		else
		{
			MessageBox.Show(this, LocalizableResources.IncompatibleBurn, MessageType.Info);
		}
	}

	/// <summary>
    /// Handles the Click event of the buttonAbout control.
    /// </summary>
    /// <param name="sender">The source of the event.</param>
    /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
	protected virtual void OnButtonAboutClicked (object sender, System.EventArgs e)
	{
		//Process.Start("http://www.marcus-furius.com/?page_id=14");
		AboutBox.Show(Gdk.Pixbuf.LoadFromResource("furiusisomount.png"));
	}

	/// <summary>
    /// Handles the Click event of the buttonEventLog control.
    /// </summary>
    /// <param name="sender">The source of the event.</param>
    /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
	protected virtual void OnButtonEventLogClicked (object sender, System.EventArgs e)
	{
		if (File.Exists(Globals.MountLog))
		{
			Process.Start(Globals.MountLog);
		}
		else
		{
			MessageBox.Show(this, string.Format(LocalizableResources.LogNotFound, 
			                                    Globals.MountLog), MessageType.Info); 
		}
	}

	/// <summary>
    /// Handles the Click event of the buttonDeleteLog control.
    /// </summary>
    /// <param name="sender">The source of the event.</param>
    /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
	protected virtual void OnButtonDeleteLogClicked (object sender, System.EventArgs e)
	{
		if (File.Exists(Globals.MountLog))
		{
			if (MessageBox.Show(this, 
		        LocalizableResources.ConfirmLogDelete, 
		        MessageType.Warning, ButtonsType.YesNo) == ResponseType.Yes)
			{
				File.Delete(Globals.MountLog);
			}
		}
	}

	/// <summary>
    /// Handles the RowActivated event of the treeviewMountedImages control.
    /// </summary>
    /// <param name="o">The object of the event.</param>
    /// <param name="args">The <see cref="Gtk.RowActivatedArgs"/> instance containing the event data.</param>
	protected virtual void OnTreeviewMountedImagesRowActivated (object o, Gtk.RowActivatedArgs args)
	{
		if (MessageBox.Show(this, 
		        LocalizableResources.RootNautilusWarning, 
		        MessageType.Warning, ButtonsType.YesNo) == ResponseType.Yes)
		{
			ImageAction.Browse(this.labelSelectedMountPoint.Text);
		}
	}

	/// <summary>
    /// Handles the Changed event of the entryImage control.
    /// </summary>
    /// <param name="sender">The source of the event.</param>
    /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
	protected virtual void OnEntryImageChanged (object sender, System.EventArgs e)
	{
		if(this.entryImage.Active == -1) // Item not on the list
		{
			return;
		}
		VerifyImage();
	}

	/// <summary>
    /// Handles the KeyRelease event of the entryImage control.
    /// </summary>
    /// <param name="o">The object of the event.</param>
    /// <param name="args">The <see cref="Gtk.KeyReleaseEventArgs"/> instance containing the event data.</param>
	protected virtual void OnEntryImageKeyReleaseEvent (object o, Gtk.KeyReleaseEventArgs args)
	{
		if (args.Event.Key == Gdk.Key.Return)
		{
			VerifyImage();	
		}
	}
	
}